// ----------------------------------------------------------------------------
/* [z-11 project]
 * (c) 1999 Copyright Malekith
 * <malekith@topnet.pl> <http://malekith.topnet.pl>
 * For info about copying see file copying.z11.
 *
 * file "Gfx.cc", created on 17/07/99.
 * descr : 
 */
// ----------------------------------------------------------------------------

#include <errno.h>
#include <fstream.h>

#include "z11.h"
#include "Gfx.h"

// ----------------------------------------------------------------------------
// globals
ColourPalette blackPalette;
Screen screen;
ColourPalette Bmp::lastImagePal, Bmp::remapTo;

// ----------------------------------------------------------------------------------
void Bitmap::load(XStr fname) {
  Bmp::load(fname);
  Bmp::remap(preserveTrans);
  if(turn) Bmp::turn();
}
// ------------------------------------------------------------------------------
void WeaponBitmap::put(MetaBmp & bmp, int nx, int ny) {
  int ww=w*bmp.w/nx,hh=h*bmp.h/ny;
  Bmp(*this,ww,hh).putS(bmp,((bmp.w-ww)>>1)+x*bmp.w/nx,bmp.h-hh);
}
// ------------------------------------------------------------------------------


// ----------------------------------------------------------------------------
// palette :
// ----------------------------------------------------------------------------
// looks for nearest (to given RGB) color. palette is 24 bit.
int ColourPalette::nearestTo(RGB what) {
  static int d8[512];
  if(d8[1]==0) { // init table to make looking for colors faster
    for(int i=0; i<512; i++)
      if(i<256) d8[i]=i*i;
      else d8[i]=(512-i)*(512-i);
  }
  int r=what.r, g=what.g, b=what.b;
  int m=maxInt;
  int mi=-1;
  RGB * p = data;
  p++;
  // note that i is initaily 1, not 0 -> no color can be maped to color 0
  // (for sprites)
  for(int i=1; i<256; i++,p++) {
    int n=d8[(r-p->r)&511]+d8[(g-p->g)&511]+d8[(b-p->b)&511];
    if(n<m) { m=n; mi=i; }
  }
  assert(mi!=-1);
  return mi;
}
// ----------------------------------------------------------------------------
void ScreenPalette::fadeOut(int msec) {
  fadeIn(msec,blackPalette);
}
// ----------------------------------------------------------------------------
void ScreenPalette::fadeIn(int msec, ColourPalette & dest) {
  timer_t t = timer();
  ColourPalette start,tmp;
  start=*this;
  msec*=ups; msec/=1000;
  while( timer() < t + msec ) {
    screen.vsync();
    
    float destm = float(clock()-t) / msec;
    float startm = 1-destm;
    for(int i=0; i<256; i++) {
      tmp[i].r = (byte) (start[i].r * startm + dest[i].r * destm);
      tmp[i].g = (byte) (start[i].g * startm + dest[i].g * destm);
      tmp[i].b = (byte) (start[i].b * startm + dest[i].b * destm);
    }
    *this = tmp;
  }
  *this = dest; // 2 B sure
}

// ----------------------------------------------------------------------------
// bitmap:
// ----------------------------------------------------------------------------
// mem alloc 
void Bmp::setSize(int nw, int nh) {
# ifdef gfxCheck
    if(nw<=0 || nh<=0) 
      haltx("nw, nh out of range %d, %d", nw, nh);
# endif
  free();
  w=nw; h=nh;
  buffer=new byte [w*h];
  line=new byte * [h];
  byte ** p = line;
  byte * x = buffer;
  while(nh-->0) {
    *p++=x;
    x+=w;
  }
}
// ----------------------------------------------------------------------------
// only pcx supported 4 now ... png in future
static XStr loadError;
void Bmp::load(XStr fname)
{
  FILE *f = fopen(fname, "rb");
  if (!f)
    throw EFile("can't open image file " + fname);
  bool ok = loadPCX(f, lastImagePal);
  fclose(f);
  if (!ok)
    throw EFile("PCX reading error! [" + loadError + "] file " + fname);
}
// ----------------------------------------------------------------------------
void Bmp::remap(bool preserveTrans) 
{
  byte ctab[256];
  for (int i = 0; i < 256; i++)
    ctab[i] = remapTo.nearestTo(lastImagePal[i]);
  int n = w * h;
  byte *ptr = buffer;
  if (preserveTrans) // color of left upper corner is transparent => maped to 0
    ctab[*ptr] = 0;  // any other color is not maped to 0
  while (n-- > 0)
    *ptr++ = ctab[*ptr];
}
// ----------------------------------------------------------------------------
// exchanges x <=> y - for texture maping
void Bmp::turn()
{
  byte *src = buffer;
  buffer = 0;
  byte **srcL = line;
  line = 0;
  setSize(h, w);
  for (int i = 0; i < w; i++)
    for (int j = 0; j < h; j++)
      line[j][i] = srcL[i][j];
  delete [] src;
  delete [] srcL;
}
// ----------------------------------------------------------------------------
// slow, c versions (fast should be in asm!)
// note full clipping support!
//
// puts entire bmp at dest at (x,y)
void Bmp::put(Bmp & dest, int x, int y) {
  if(x>=dest.w || y>=dest.h || x+w<0 || y+w<0) return; // totally outside
  int srcLine=0;
  if(y<0) { srcLine=-y; y=0; }
  if(x<0) {
    int srcAdd=-x;
    int lineWidth=w-srcAdd;
    while(srcLine<h && y<dest.h) 
      memcpy(dest.line[y++],line[srcLine++]+srcAdd,lineWidth);
  }
  else {
    int lineWidth = dest.w-x <? w;
    while(srcLine<h && y<dest.h) 
      memcpy(dest.line[y++]+x,line[srcLine++],lineWidth);
  }
}
inline void memcpy0(byte * dst, byte * src, int n) {
  while(n-->0) {
    if(*src) *dst=*src;
    src++; dst++;
  }
}
// ----------------------------------------------------------------------------
void Bmp::putS(Bmp & dest, int x, int y) {
  if(x>=dest.w || y>=dest.h || x+w<0 || y+w<0) return; // totally outside
  int srcLine=0;
  if(y<0) { srcLine=-y; y=0; }
  if(x<0) {
    int srcAdd=-x;
    int lineWidth=w-srcAdd;
    while(srcLine<h && y<dest.h) 
      memcpy0(dest.line[y++],line[srcLine++]+srcAdd,lineWidth);
  }
  else {
    int lineWidth = dest.w-x <? w;
    while(srcLine<h && y<dest.h) 
      memcpy0(dest.line[y++]+x,line[srcLine++],lineWidth);
  }
}
// ----------------------------------------------------------------------------
// on svgalib -> don't even think about gl_scalebox -- for it context must
// be set !
void resizeBox(int x1, int y1, void * s, int x2, int y2, void * d) {
  if(x1==x2 && y1==y2) { memcpy(d,s,x1*y1); return; }
  int dx=(x1<<16)/x2;
  int dy=(y1<<16)/y2;
  int sy=0;
  register byte * dst = (byte*)d;
  while(y2-->0) {
    register byte * src = (byte*)s+(sy>>16)*x1;
    sy+=dy;
    register int sx=0;
    register int n=x2;
    while(n-->0) { *dst++=src[sx>>16]; sx+=dx; }
  }
}
// ----------------------------------------------------------------------------
void Bmp::resize(int dw, int dh) {
  byte * src = buffer;
  buffer=0; // it's little bit ugly ... 
  int sw=w, sh=h;
  setSize(dw,dh);
  resizeBox(sw,sh,src,dw,dh,buffer);
  delete [] src;
}
// ----------------------------------------------------------------------------
Bmp::Bmp(Bmp & src, int ww, int hh) : lockedMem(0), buffer(0), line(0) {
  setSize(ww,hh);
  resizeBox(src.w,src.h,src.buffer,w,h,buffer);
}
// ----------------------------------------------------------------------------
void Bmp::set(int color) {
  memset(buffer,color,w*h);
}
// ----------------------------------------------------------------------------
void Bmp::copy(Bmp & src) {
  if(!buffer || src.w!=w || src.h!=h)
    setSize(src.w,src.h);
  memcpy(buffer,src.buffer,w*h);
}
// ----------------------------------------------------------------------------


// ----------------------------------------------------------------------------
void Bmp::plot(int x, int y, int color) {
  if(x<0 || y<0 || x>=w || y>=h) return; // clipping must be done ...
  line[y][x]=color;
}
// ----------------------------------------------------------------------------
// i *don't* think it's fast ...
void Bmp::draw(int x1, int y1, int x2, int y2, int color) {
  if(x1==x2 && y1==y2) { plot(x1,y1,color); return; }
  if(abs(x1-x2)>abs(y1-y2)) {
    if(x1>x2) { swap(x1,x2); swap(y1,y2); } 
    int d=((y2-y1)<<16)/(x2-x1);
    y1<<=16;
    int mh=h<<16;
    while(x1<=x2) {
      if(x1>=0 && y1>=0 && x1<w && y1<mh) line[y1>>16][x1]=color;
      y1+=d;
      x1++;
    }
  }
  else {
    if(y1>y2) { swap(x1,x2); swap(y1,y2); } 
    int d=((x2-x1)<<16)/(y2-y1);
    x1<<=16;
    int mw=w<<16;
    while(y1<=y2) {
      if(x1>=0 && y1>=0 && x1<mw && y1<h) line[y1][x1>>16]=color;
      x1+=d;
      y1++;
    }
  }
}
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// actual file reading
// ----------------------------------------------------------------------------
// Reads a 16 bit word from a file, using intel byte ordering (little endian)
// this *will* also on motorola (big endian) i.e. it will read intel
// ordered bytes -- don't need to change
int igetw(FILE *f) {
   int b1, b2;
   if ( (b1 = getc(f)) != EOF && (b2 = getc(f)) != EOF )
	 return ((b2 << 8) | b1);
   return EOF;
}
// ----------------------------------------------------------------------------
int igetl(FILE *f) {
   int b1, b2, b3, b4;

   if ((b1 = getc(f)) != EOF && (b2 = getc(f)) != EOF &&
       (b3 = getc(f)) != EOF && (b4 = getc(f)) != EOF )
	       return (((long)b4 << 24) | ((long)b3 << 16) |
		       ((long)b2 << 8) | (long)b1);
   return EOF;
}
// ----------------------------------------------------------------------------
// striped Allegro load_pcx routine
#define stopLoading(a) ({ loadError=a; return false; })
bool Bmp::loadPCX(FILE * f, ColourPalette & pal)
{
  int c;
  int width, height;
  int bpp, bytes_per_line;
  int xx;
  int x, y;
  char ch;

  getc(f);			/* skip manufacturer ID */
  getc(f);			/* skip version flag */
  getc(f);			/* skip encoding flag */

  if (getc(f) != 8)		/* we like 8 bit color planes */
    stopLoading("not 8 bpp");

  width = -(igetw(f));		/* xmin */
  height = -(igetw(f));		/* ymin */
  width += igetw(f) + 1;	/* xmax */
  height += igetw(f) + 1;	/* ymax */

  igetl(f);			/* skip DPI values */

  for (c = 0; c < 16; c++) {	/* read the 16 color pallete */
    pal[c].r = getc(f);
    pal[c].g = getc(f);
    pal[c].b = getc(f);
  }

  getc(f);

  bpp = getc(f) * 8;		/* how many color planes? */

  if (bpp != 8)
    stopLoading("not 8 bpp [2]");

  bytes_per_line = igetw(f);

  for (c = 0; c < 60; c++)	// skip some more junk 
    getc(f);

  setSize(width, height);

  for (y = 0; y < height; y++) {	// read RLE encoded PCX data 
    x = xx = 0;

    while (x < bytes_per_line * bpp / 8) {
      ch = getc(f);
      if ((ch & 0xC0) == 0xC0) {
	c = (ch & 0x3F);
	ch = getc(f);
      }
      else
	c = 1;

      while (c--) {
	if (x < width)
	  line[y][x] = ch;
	x++;
      }
    }
  }
  while (!feof(f)) {
    if (getc(f) == 12) {
      for (c = 0; c < 256; c++) {
	pal[c].r = getc(f);
	pal[c].g = getc(f);
	pal[c].b = getc(f);
      }
      break;
    }
  }
//if (errno) stopLoading(xprintf("errno ocured (%d)",errno));

  return true;
}

// ----------------------------------------------------------------------------
// Font
// ----------------------------------------------------------------------------
void Font::load(XStr fname) {
  delete [] buffer; delete [] chr;
  ifstream s; s.open(fname);
  if(!s) throw EFile("can't open font file "+fname);
  byte sizes[2];
  s.read(sizes,sizeof(sizes));
  w=sizes[0]; h=sizes[1];
  if(w>8 || h>16) throw EFile("bad font file format. ("+fname+")");
  s.seekg(0,ios::end);
  int len=s.tellg()-sizeof(sizes);
  s.seekg(sizeof(sizes));
  maxChr=len/16;
  if(maxChr*16!=len) throw EFile("bad font file length. ("+fname+")");
  buffer=new byte[maxChr*w*h];
  chr=new byte*[maxChr];
  byte * p = buffer;
  for(int i=0; i<maxChr; i++) {
    chr[i]=p;
    for(int y=0; y<h; y++) {
      if(!s) haltx("font read error! ("+fname+")"); // ???
      byte d;
      s.read(&d,sizeof(d));
      int m=0x80;      
      for(int x=0; x<w; x++,m>>=1)
        if(m&d)  *p++=rand()%2+1;
	else     *p++=0;
    }
    if(16-h>0) s.ignore(16-h);
  }
}
// ----------------------------------------------------------------------------
int * fct(RGB c1, float d) {
  static int colors[3];
  static int oldColor=-1;
  static float oldD=-1;
  if(oldColor==*(int*)&c1 && d==oldD) return colors;
  colors[1]=screen.palette.nearestTo(c1);
  colors[2]=screen.palette.nearestTo(((RGB){ (byte)(c1.r*d), 
					     (byte)(c1.g*d), 
					     (byte)(c1.b*d) }));
  oldD=d;
  oldColor=*(int*)&c1;
  return colors;
}
// ----------------------------------------------------------------------------
Bmp::Bmp(XStr string, Font & font, int color, int attr, float d) 
  : lockedMem(0), buffer(0), line(0) {
  // colors setting
  static int oldColor=-1;
  static float oldD=-1;
  static byte colors[3];
  if(oldColor!=color || oldD!=d) {
    oldColor=color; oldD=d;
    colors[0]=0;
    RGB c = { (byte)(color>>16), (byte)(color>>8), (byte)(color) };
    colors[1]=(byte)screen.palette.nearestTo(c);
    colors[2]=(byte)screen.palette.nearestTo(((RGB){ (byte)(c.r*d),
                                                     (byte)(c.g*d), 
						     (byte)(c.b*d) }));
  }
  // set size
  setSize(string.length()*font.w,font.h);
  char * p = string;
  int add = 0;
  while(*p) {
    byte * f = font.chr[*p++];
/*    
XStr a=XStr("c(")+*p+"): ";
for(int i=0; i<font.w*font.h; i++) if(f[i]) a+="#"; else a+=" ";
dm("%s",a.chars());
*/    
    for(int i=0; i<h; i++) {
      byte * v = line[i]+add;
      int n=font.w;
      if( (attr & Font::underline) && i==h-2 ) 
	while(n-->0) *v++=colors[(*f++&1)+1];
      else if( (attr & Font::bold) ) {
        n--; *v++=colors[*f++];
        while(n-->0) *v++=colors[(*(f-1)) >? (*f++)];
      }
      else
        while(n-->0) *v++=colors[*f++];
    }
    add+=font.w;
  } 
}
// ----------------------------------------------------------------------------


