// ----------------------------------------------------------------------------------
/* [z-11 project]
 * (c) 1999 Copyright Malekith
 * <malekith@topnet.pl> <http://malekith.topnet.pl>
 * For info about copying see file copying.z11.
 *
 * file "Xcpt.cc", created on 14/07/99.
 * descr : debug stuff
 */
// ----------------------------------------------------------------------------------

#ifdef DJGPP
#  include <signal.h>
#  include <setjmp.h>
#  include <sys/exceptn.h>
#endif

#include <errno.h>

#include "z11.h"
#include "Gfx.h"
#include "Input.h"

FILEINFO(Xcpt);

// ----------------------------------------------------------------------------------
// G L O B A L S

FILE * dbg;
bool debug=true;

// ----------------------------------------------------------------------------------
// error handling
EDead deadFoodErrno(XStr msg, char * file, int line, char * func) {
  const char * ed = errno < sys_nerr ? 
                    sys_errlist[errno] : 
		    "errno out of range!";
  EDead e(xprintf("%s\nerrno: %s (%d)\nLoc: %s [%d] / %s",
          msg.chars(),
	  ed,errno,
	  file,line,func));
#  ifdef DSX
    throw e;
#  else
    handleException(e);
#  endif
}
// ----------------------------------------------------------------------------------
EDead deadFood(XStr msg, char * file, int line, char * func) {
  EDead e(xprintf("%s\nLoc: %s [%d] / %s",msg.chars(),file,line,func));
#  ifdef DSX
    throw e;
#  else
    handleException(e);
#  endif
}
// ----------------------------------------------------------------------------------
void deadAssert(char * test, char * file, int line, char * func) {
  EAssert e(xprintf("Assertion failed %s\nLoc: %s [%d] / %s",test,file,line,func));
# ifdef DSX
    throw e;
#  else
    handleException(e);
#  endif
}
// ----------------------------------------------------------------------------------
XStr EBase::format() {
  return xprintf("Xcpt: %s: '%s' <%s>\n",id().chars(),xmsg.chars(),version);
}
// ----------------------------------------------------------------------------------
#ifdef DJGPP
static jmp_buf abort_jmp;
#endif

void handleException(EBase & e) {
  kbd->kill();
  screen.kill();
  fprintf(stderr,
          "[z-11] has raised unhandled exception '%s', with message:\n"
          "%s\n",e.id().chars(),e.xmsg.chars());

  if(debug) {
    fprintf(debugFile(),"dead: "+e.format());
    fclose(dbg);
  }
  
# ifdef DJGPP
  if(!setjmp(abort_jmp)) {
    __djgpp_exception_state_ptr = &abort_jmp;
    raise(SIGABRT);
  }
# endif

  _exit(1);
  for(;;){} // - to disable warning, that this function returns ...
}
// ----------------------------------------------------------------------------------
FILE * debugFile() {
  if(dbg) return dbg;
  dbg=fopen(txtData+"debuginf.z11","at");
  time_t now;
  time(&now);
  tm t = *localtime(&now);
  setbuf(dbg,0);
  fprintf(dbg,"*** [%s] run on %d/%02d/%02d at %d:%02d:%02d\n",version,t.tm_year,t.tm_mon+1,t.tm_mday,t.tm_hour,t.tm_min,t.tm_sec);
  return dbg;
}
// ----------------------------------------------------------------------------------
// overloaded global new/delete
void * operator new (size_t s) {
  void * p = malloc(s);
  if(s && !p) {
    screen.kill();
    fprintf(stderr,
      "I'm very sorry - out of memory. If you have over 16MB it's bug of\n"
      "this program. If not - it still should be consider as a bug (unless\n"
      "you have about 1MB...) Plaese report to author.\n"
      "Now (actually, after 5 seconds) normal exit procedure will be performed\n"
      "but it may fail.\n");
    rest(5000);
    throw EDead("Out of memory");
  }
  return p;
}
// ----------------------------------------------------------------------------------
void operator delete(void * ptr) {
  free(ptr);
}
// ----------------------------------------------------------------------------------
void my2ArgErrorHandler(const char * m1, const char * m2) {
  EDead e(xprintf("%s: %s",m1,m2));
# ifdef DSX
    throw e;
# else
    handleException(e);
# endif
}
// ----------------------------------------------------------------------------------



