// ----------------------------------------------------------------------------
/* [z-11 project]
 * (c) 1999 Copyright Malekith
 * <malekith@topnet.pl> <http://malekith.topnet.pl>
 * For info about copying see file copying.z11.
 *
 * file "Gfx.h", created on 17/07/99.
 * descr :  go figure
 */
// ----------------------------------------------------------------------------

#ifndef GfxH
#define GfxH

// ----------------------------------------------------------------------------
#ifdef gfxCheck
#define Check(x) assert(x)
#else
#define Check(x) ((void)0)
#endif
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
struct RGB { 
  byte r,g,b; 
}; 
// r,g,b 8 bits ! (not 6)
// ----------------------------------------------------------------------------
class ColourPalette {
  protected:
    RGB data[256];
  public:
    void * getData() { return data; }
    RGB & operator[](int i) { return data[i]; }    
    virtual void set(int i, RGB r) { data[i]=r; }
    ColourPalette & operator =(ColourPalette & p) 
      { memcpy(data,p.data,sizeof(data)); return * this; }
    int nearestTo(RGB what);
};
// ----------------------------------------------------------------------------
class Screen;
class ScreenPalette : public ColourPalette {
    friend class Screen;
    ScreenPalette() { } // priv ctor -> only TScreen can make this object
  public:
    const RGB & operator[](int i) { return data[i]; }    
    virtual void set(int i, RGB r);
    ScreenPalette & operator =(ColourPalette & p);
    void fadeIn(int msec, ColourPalette & dest);
    void fadeOut(int msec);
};
extern ColourPalette blackPalette;
// ----------------------------------------------------------------------------
// base class for screen and bmp
class MetaBmp {
  public:
    int w,h;
    MetaBmp() { }
    virtual bool isScreen() = 0;
    virtual void plot(int x, int y, int color) = 0;
    virtual void draw(int x1, int y1, int x2, int y2, int color) = 0;
    virtual void set(int c) = 0;
    void clear() { set(0); }
    virtual ~MetaBmp() { }
};
// ----------------------------------------------------------------------------
class Screen : public MetaBmp {
  public:
    ScreenPalette palette;
    
    void init(); // if can't set throws EScreen xcpt
    void kill(); // switch back to TEXT mode
    void vsync();
    void setRes(int resX, int resY); // didto
    virtual bool isScreen() { return true; }
    virtual void plot(int x, int y, int color);
    virtual void draw(int x1, int y1, int x2, int y2, int color);
    virtual void set(int color);
}; 
extern Screen screen; 
// ----------------------------------------------------------------------------
inline int color(int r, int g, int b) { 
  return screen.palette.nearestTo( ((RGB) { (byte)r, (byte)g, (byte)b }) );
}
// ----------------------------------------------------------------------------
inline int color(int c) {
  return color((c>>16)&0xff,(c>>8)&0xff,c&0xff);
}
// ----------------------------------------------------------------------------
class Font {
  public:
    enum { 
           normal    = 0x00000000, 
           bold      = 0x00000001,
	   underline = 0x00000002,
	   italic    = 0x00000004 // ignored
         }; 
    int w,h;
    int maxChr;
    byte ** chr; // point to array of ptrs to buffer
    byte * buffer; // points to w*h*maxCharacter buffer
    Font() : chr(0), buffer(0) { }
    ~Font() { delete [] chr; delete [] buffer; }
    void load(XStr fname);
};
// ----------------------------------------------------------------------------
class Bmp : public MetaBmp {
    bool loadPCX(FILE * f, ColourPalette & pal);
    void copy(Bmp & src); // makes exact copy of src. new mem is allocated
    void * lockedMem;
  public:
    byte * buffer;
    byte ** line;
    static ColourPalette lastImagePal; // set by load funcs
    static ColourPalette remapTo; // palette used by remap() method
    
    Bmp() : lockedMem(0), buffer(0), line(0) { }
    Bmp(int ww, int hh) : lockedMem(0), buffer(0), line(0) { setSize(ww,hh); }
    Bmp(Bmp & src)     : lockedMem(0), buffer(0), line(0) { copy(src); }
    Bmp(Bmp & src, int ww, int hh); // initializes with resized src
    Bmp(XStr string, Font & font, int color = 0xffffff, 
         int attr = Font::normal, float colorDiff = 0.7); 
	 // initializes with printed string
    ~Bmp() { free(); }
    
    Bmp & operator = (Bmp & src) { copy(src); return *this; }    
    
    // put, get from screen
    void put(int x, int y); // on screen
    void get(int x, int y); // from screen
    void putS(int x, int y); // color 0 transculent (S for Sprite) ...
    
    // copy this bitmap to another. clipping is done.
    void put(Bmp & dest, int x, int y); // on dest
    void putS(Bmp & dest, int x, int y); // on dest, mask color 0
    void get(Bmp & src, int x, int y) { src.put(*this,-x,-y); }
    // image manipulation
    void load(XStr filename); // from file. EFile except could be thrown
    void resize(int dw, int dh);
    void turn();
    void remap(bool preserveTrans);
    virtual void set(int color); // sets to color
    
    // memory manipulation
    void setSize(int nw, int nh);
    void free(); // frees memory and clears ptrs
    // mem protection
    void lock(); // marks bmp read-only
    bool locked() { return lockedMem; }
    
    // MetaBmp support    
    bool isScreen() { return false; }
    void put(MetaBmp & src, int x, int y) {
      if(src.isScreen()) put(x,y);
      else put(*(Bmp*)&src,x,y);
    }
    void putS(MetaBmp & src, int x, int y) {
      if(src.isScreen()) putS(x,y);
      else putS(*(Bmp*)&src,x,y);
    }
    virtual void plot(int x, int y, int color);
    virtual void draw(int x1, int y1, int x2, int y2, int color);
};
// ----------------------------------------------------------------------------


// ----------------------------------------------------------------------------------
// bmp for sprites
// at this point AddUser & RemoveUser funcs arn't called ...
class Bitmap : public Bmp {
  public:
    OkStuff(0xb14a9ff);
    XStr name;
    bool preserveTrans,turn;
    
    Bitmap() : preserveTrans(false), turn(false) { setId(); }

    void load(XStr fname); // if failed EFile exception is thrown
};
// ----------------------------------------------------------------------------------
class WallBitmap : public Bitmap {
  public:
    int x,y;
    WallBitmap() { preserveTrans=false; turn=true; }
    void RenderSliver(int scale, int column, int lightDistance);
};
// ----------------------------------------------------------------------------------
class SpriteBitmap : public WallBitmap {
  public:
    SpriteBitmap() { preserveTrans=true; turn=true; }
    void RenderSliver(int scale, int column, int distance);
};
// ----------------------------------------------------------------------------------
class WeaponBitmap : public Bitmap {
  public:
    int x,y;
    WeaponBitmap() : x(0), y(0) { preserveTrans=true; turn=false; }
    void put(MetaBmp & target, int nx, int ny);
    // puts at the bottom of the screen ('where')
    // nx,ny are window dimensions to which scaling isn't needed
};
// ----------------------------------------------------------------------------------
extern Bmp doubleBuffer;
// ----------------------------------------------------------------------------------

#endif // GfxH



