// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

  this file has been modified by Malekith (malekith@topnet.pl)
  XStr class name has been changed to XStr to avoid name confilicts
  there are no limits now for length of string (in fact there is - 2GB)
  operator (const char*) has been changed to (char*) because i don't like const

*/


#ifndef XStrH

#define XStrH

#include <iostream.h>


#undef OK

// ------------------- from builtin.h ------------------
typedef void (*two_arg_error_handler_t)(const char*, const char*);
extern two_arg_error_handler_t lib_error_handler;

// ---------------- Regex.h --------------------------
#if defined(SHORT_NAMES) || defined(VMS)
#define re_compile_pattern	recmppat
#define re_pattern_buffer	repatbuf
#define re_registers		reregs
#endif

struct re_pattern_buffer;       // defined elsewhere
struct re_registers;

class Regex
{
private:

                     Regex(const Regex&) {}  // no X(X&)
  void               operator = (const Regex&) {} // no assignment

protected:
  re_pattern_buffer* buf;
  re_registers*      reg;

public:
                     Regex(const char* t, 
                           int fast = 0, 
                           int bufsize = 40, 
                           const char* transtable = 0);

                    ~Regex();

  int                match(const char* s, int len, int pos = 0) const;
  int                search(const char* s, int len, 
                            int& matchlen, int startpos = 0) const;
  int                match_info(int& start, int& length, int nth = 0) const;

  int                OK() const;  // representation invariant
};

// some built in regular expressions

extern const Regex RXwhite;          // = "[ \n\t\r\v\f]+"
extern const Regex RXint;            // = "-?[0-9]+"
extern const Regex RXdouble;         // = "-?\\(\\([0-9]+\\.[0-9]*\\)\\|
                                     //    \\([0-9]+\\)\\|\\(\\.[0-9]+\\)\\)
                                     //    \\([eE][---+]?[0-9]+\\)?"
extern const Regex RXalpha;          // = "[A-Za-z]+"
extern const Regex RXlowercase;      // = "[a-z]+"
extern const Regex RXuppercase;      // = "[A-Z]+"
extern const Regex RXalphanum;       // = "[0-9A-Za-z]+"
extern const Regex RXidentifier;     // = "[A-Za-z_][A-Za-z0-9_]*"
// -------------------- eof Regex.h --------------------------


struct XStrRep                     // internal XStr representations
{
  int    len;         // string length
  int    sz;          // allocated space
  char   s[1];        // the string starts here
                      // (at least 1 char for trailing null)
                      // allocated & expanded via non-public fcts
};

// primitive ops on StrReps -- nearly all XStr fns go through these.

XStrRep*     Salloc(XStrRep*, const char*, int, int);
XStrRep*     Scopy(XStrRep*, const XStrRep*);
XStrRep*     Scat(XStrRep*, const char*, int, const char*, int);
XStrRep*     Scat(XStrRep*, const char*, int,const char*,int, const char*,int);
XStrRep*     Sprepend(XStrRep*, const char*, int);
XStrRep*     Sreverse(const XStrRep*, XStrRep*);
XStrRep*     Supcase(const XStrRep*, XStrRep*);
XStrRep*     Sdowncase(const XStrRep*, XStrRep*);
XStrRep*     Scapitalize(const XStrRep*, XStrRep*);

// These classes need to be defined in the order given

class XStr;
class XSubStr;

class XSubStr
{
  friend class      XStr;
protected:

  XStr&           S;        // The XStr I'm a substring of
  int    pos;      // starting position in S's rep
  int    len;      // length of substring

  void              assign(const XStrRep*, const char*, int = -1);
                    XSubStr(XStr& x, int p, int l);
                    XSubStr(const XSubStr& x);

public:

// Note there are no public constructors. SubStrings are always
// created via XStr operations

                   ~XSubStr();

  XSubStr&        operator =  (const XStr&     y);
  XSubStr&        operator =  (const XSubStr&  y);
  XSubStr&        operator =  (const char* t);
  XSubStr&        operator =  (char        c);

// return 1 if target appears anywhere in XSubStr; else 0

  int               contains(char        c) const;
  int               contains(const XStr&     y) const;
  int               contains(const XSubStr&  y) const;
  int               contains(const char* t) const;
  int               contains(const Regex&       r) const;

// return 1 if target matches entire XSubStr

  int               matches(const Regex&  r) const;

// IO 

  friend ostream&   operator<<(ostream& s, const XSubStr& x);

// status

  int               length() const;
  int               empty() const;

  const char*       chars() const;
  const char*       cChars() const;

  int               OK() const; 

};


class XStr
{
  friend class      XSubStr;

protected:
  XStrRep*           rep;   // Strings are pointers to their representations

// some helper functions

  int               search(int, int, const char*, int = -1) const;
  int               search(int, int, char) const;
  int               match(int, int, int, const char*, int = -1) const;
  int               _gsub(const char*, int, const char* ,int);
  int               _gsub(const Regex&, const char*, int);
  XSubStr           _substr(int, int);

public:

// constructors & assignment

                    XStr();
                    XStr(const XStr& x);
                    XStr(const XSubStr&  x);
                    XStr(const char* t);
                    XStr(const char* t, int len);
                    XStr(char c);

                    ~XStr();

  XStr&           operator =  (const XStr&     y);
  XStr&           operator =  (const char* y);
  XStr&           operator =  (char        c);
  XStr&           operator =  (const XSubStr&  y);

// concatenation

  XStr&           operator += (const XStr&     y); 
  XStr&           operator += (const XSubStr&  y);
  XStr&           operator += (const char* t);
  XStr&           operator += (char        c);

  void              prepend(const XStr&     y); 
  void              prepend(const XSubStr&  y);
  void              prepend(const char* t);
  void              prepend(char        c);


// procedural versions:
// concatenate first 2 args, store result in last arg

  friend inline void     cat(const XStr&, const XStr&, XStr&);
  friend inline void     cat(const XStr&, const XSubStr&, XStr&);
  friend inline void     cat(const XStr&, const char*, XStr&);
  friend inline void     cat(const XStr&, char, XStr&);

  friend inline void     cat(const XSubStr&, const XStr&, XStr&);
  friend inline void     cat(const XSubStr&, const XSubStr&, XStr&);
  friend inline void     cat(const XSubStr&, const char*, XStr&);
  friend inline void     cat(const XSubStr&, char, XStr&);

  friend inline void     cat(const char*, const XStr&, XStr&);
  friend inline void     cat(const char*, const XSubStr&, XStr&);
  friend inline void     cat(const char*, const char*, XStr&);
  friend inline void     cat(const char*, char, XStr&);

// double concatenation, by request. (yes, there are too many versions, 
// but if one is supported, then the others should be too...)
// Concatenate first 3 args, store in last arg

  friend inline void     cat(const XStr&,const XStr&, const XStr&,XStr&);
  friend inline void     cat(const XStr&,const XStr&,const XSubStr&,XStr&);
  friend inline void     cat(const XStr&,const XStr&, const char*, XStr&);
  friend inline void     cat(const XStr&,const XStr&, char, XStr&);
  friend inline void     cat(const XStr&,const XSubStr&,const XStr&,XStr&);
  inline friend void     cat(const XStr&,const XSubStr&,const XSubStr&,XStr&);
  friend inline void     cat(const XStr&,const XSubStr&, const char*, XStr&);
  friend inline void     cat(const XStr&,const XSubStr&, char, XStr&);
  friend inline void     cat(const XStr&,const char*, const XStr&,    XStr&);
  friend inline void     cat(const XStr&,const char*, const XSubStr&, XStr&);
  friend inline void     cat(const XStr&,const char*, const char*, XStr&);
  friend inline void     cat(const XStr&,const char*, char, XStr&);

  friend inline void     cat(const char*, const XStr&, const XStr&,XStr&);
  friend inline void     cat(const char*,const XStr&,const XSubStr&,XStr&);
  friend inline void     cat(const char*,const XStr&, const char*, XStr&);
  friend inline void     cat(const char*,const XStr&, char, XStr&);
  friend inline void     cat(const char*,const XSubStr&,const XStr&,XStr&);
  friend inline void     cat(const char*,const XSubStr&,const XSubStr&,XStr&);
  friend inline void     cat(const char*,const XSubStr&, const char*, XStr&);
  friend inline void     cat(const char*,const XSubStr&, char, XStr&);
  friend inline void     cat(const char*,const char*, const XStr&,    XStr&);
  friend inline void     cat(const char*,const char*, const XSubStr&, XStr&);
  friend inline void     cat(const char*,const char*, const char*, XStr&);
  friend inline void     cat(const char*,const char*, char, XStr&);


// searching & matching

// return position of target in string or -1 for failure

  int               index(char        c, int startpos = 0) const;      
  int               index(const XStr&     y, int startpos = 0) const;      
  int               index(const XSubStr&  y, int startpos = 0) const;      
  int               index(const char* t, int startpos = 0) const;  
  int               index(const Regex&      r, int startpos = 0) const;       

// return 1 if target appears anyhere in XStr; else 0

  int               contains(char        c) const;
  int               contains(const XStr&     y) const;
  int               contains(const XSubStr&  y) const;
  int               contains(const char* t) const;
  int               contains(const Regex&      r) const;

// return 1 if target appears anywhere after position pos 
// (or before, if pos is negative) in XStr; else 0

  int               contains(char        c, int pos) const;
  int               contains(const XStr&     y, int pos) const;
  int               contains(const XSubStr&  y, int pos) const;
  int               contains(const char* t, int pos) const;
  int               contains(const Regex&      r, int pos) const;

// return 1 if target appears at position pos in XStr; else 0

  int               matches(char        c, int pos = 0) const;
  int               matches(const XStr&     y, int pos = 0) const;
  int               matches(const XSubStr&  y, int pos = 0) const;
  int               matches(const char* t, int pos = 0) const;
  int               matches(const Regex&      r, int pos = 0) const;

//  return number of occurences of target in XStr

  int               freq(char        c) const; 
  int               freq(const XStr&     y) const;
  int               freq(const XSubStr&  y) const;
  int               freq(const char* t) const;

// XSubStr extraction

// Note that you can't take a substring of a const XStr, since
// this leaves open the possiblility of indirectly modifying the
// XStr through the XSubStr

  XSubStr         at(int         pos, int len);
  XSubStr         operator () (int         pos, int len); // synonym for at

  XSubStr         at(const XStr&     x, int startpos = 0); 
  XSubStr         at(const XSubStr&  x, int startpos = 0); 
  XSubStr         at(const char* t, int startpos = 0);
  XSubStr         at(char        c, int startpos = 0);
  XSubStr         at(const Regex&      r, int startpos = 0); 

  XSubStr         before(int          pos);
  XSubStr         before(const XStr&      x, int startpos = 0);
  XSubStr         before(const XSubStr&   x, int startpos = 0);
  XSubStr         before(const char*  t, int startpos = 0);
  XSubStr         before(char         c, int startpos = 0);
  XSubStr         before(const Regex&       r, int startpos = 0);

  XSubStr         through(int          pos);
  XSubStr         through(const XStr&      x, int startpos = 0);
  XSubStr         through(const XSubStr&   x, int startpos = 0);
  XSubStr         through(const char*  t, int startpos = 0);
  XSubStr         through(char         c, int startpos = 0);
  XSubStr         through(const Regex&       r, int startpos = 0);

  XSubStr         from(int          pos);
  XSubStr         from(const XStr&      x, int startpos = 0);
  XSubStr         from(const XSubStr&   x, int startpos = 0);
  XSubStr         from(const char*  t, int startpos = 0);
  XSubStr         from(char         c, int startpos = 0);
  XSubStr         from(const Regex&       r, int startpos = 0);

  XSubStr         after(int         pos);
  XSubStr         after(const XStr&     x, int startpos = 0);
  XSubStr         after(const XSubStr&  x, int startpos = 0);
  XSubStr         after(const char* t, int startpos = 0);
  XSubStr         after(char        c, int startpos = 0);
  XSubStr         after(const Regex&      r, int startpos = 0);


// deletion

// delete len chars starting at pos
  void              del(int         pos, int len);

// delete the first occurrence of target after startpos

  void              del(const XStr&     y, int startpos = 0);
  void              del(const XSubStr&  y, int startpos = 0);
  void              del(const char* t, int startpos = 0);
  void              del(char        c, int startpos = 0);
  void              del(const Regex&      r, int startpos = 0);

// global substitution: substitute all occurrences of pat with repl

  int               gsub(const XStr&     pat, const XStr&     repl);
  int               gsub(const XSubStr&  pat, const XStr&     repl);
  int               gsub(const char* pat, const XStr&     repl);
  int               gsub(const char* pat, const char* repl);
  int               gsub(const Regex&      pat, const XStr&     repl);

// friends & utilities

// split string into array res at separators; return number of elements

  friend int        split(const XStr& x, XStr res[], int maxn, 
                          const XStr& sep);
  friend int        split(const XStr& x, XStr res[], int maxn, 
                          const Regex&  sep);

  friend XStr     common_prefix(const XStr& x, const XStr& y, 
                                  int startpos = 0);
  friend XStr     common_suffix(const XStr& x, const XStr& y, 
                                  int startpos = -1);
  friend XStr     replicate(char        c, int n);
  friend XStr     replicate(const XStr&     y, int n);
  friend XStr     join(XStr src[], int n, const XStr& sep);

// simple builtin transformations

  friend inline XStr     reverse(const XStr& x);
  friend inline XStr     upcase(const XStr& x);
  friend inline XStr     downcase(const XStr& x);
  friend inline XStr     capitalize(const XStr& x);

// in-place versions of above

  void              reverse();
  void              upcase();
  void              downcase();
  void              capitalize();

// element extraction

  char&             operator [] (int i);
  const char&       operator [] (int i) const;
  char              elem(int i) const;
  char              firstchar() const;
  char              lastchar() const;

// conversion

  const char * cChars() const;

  // changed by Malekith
  char*             chars() { return rep->s; }
                    operator char*() { return chars(); }
  // added by Malekith
  int asInt(); // also supports 0x, 0
  float asFloat() { return atof(chars()); }
  
// IO

  friend inline ostream&   operator<<(ostream& s, const XStr& x);
  friend ostream&   operator<<(ostream& s, const XSubStr& x);
  friend istream&   operator>>(istream& s, XStr& x);

  friend int        readline(istream& s, XStr& x, 
                             char terminator = '\n',
                             int discard_terminator = 1);

// status

  int               length() const;
  int               empty() const;

// preallocate some space for XStr
  void              alloc(int newsize);

// report current allocation (not length!)

  int               allocation() const;


  void     error(const char* msg) const;

  int               OK() const;
};

typedef XStr StrTmp; // for backward compatibility

// other externs

int        compare(const XStr&    x, const XStr&     y);
int        compare(const XStr&    x, const XSubStr&  y);
int        compare(const XStr&    x, const char* y);
int        compare(const XSubStr& x, const XStr&     y);
int        compare(const XSubStr& x, const XSubStr&  y);
int        compare(const XSubStr& x, const char* y);
int        fcompare(const XStr&   x, const XStr&     y); // ignore case

extern XStrRep  _nilStrRep;
extern XStr _nilString;

// status reports, needed before defining other things

inline int         XStr::length() const {  return rep->len; }
inline int         XStr::empty() const { return rep->len == 0; }
inline const char * XStr::cChars() const { return &(rep->s[0]); }
inline int         XStr::allocation() const { return rep->sz; }

inline int         XSubStr::length() const { return len; }
inline int         XSubStr::empty() const { return len == 0; }
inline const char* XSubStr::chars() const { return &(S.rep->s[pos]); }
inline const char* XSubStr::cChars() const { return &(S.rep->s[pos]); }

// constructors

inline XStr::XStr() 
  : rep(&_nilStrRep) {}
inline XStr::XStr(const XStr& x) 
  : rep(Scopy(0, x.rep)) {}
inline XStr::XStr(const char* t) 
  : rep(Salloc(0, t, -1, -1)) {}
inline XStr::XStr(const char* t, int tlen)
  : rep(Salloc(0, t, tlen, tlen)) {}
inline XStr::XStr(const XSubStr& y)
  : rep(Salloc(0, y.cChars(), y.length(), y.length())) {}
inline XStr::XStr(char c) 
  : rep(Salloc(0, &c, 1, 1)) {}

inline XStr::~XStr() { if (rep != &_nilStrRep) delete rep; }

inline XSubStr::XSubStr(const XSubStr& x)
  :S(x.S), pos(x.pos), len(x.len) {}
inline XSubStr::XSubStr(XStr& x, int first, int l)
  :S(x), pos(first), len(l) {}

inline XSubStr::~XSubStr() {}

// assignment

inline XStr& XStr::operator =  (const XStr& y)
{ 
  rep = Scopy(rep, y.rep);
  return *this;
}

inline XStr& XStr::operator=(const char* t)
{
  rep = Salloc(rep, t, -1, -1);
  return *this;
}

inline XStr& XStr::operator=(const XSubStr&  y)
{
  rep = Salloc(rep, y.cChars(), y.length(), y.length());
  return *this;
}

inline XStr& XStr::operator=(char c)
{
  rep = Salloc(rep, &c, 1, 1);
  return *this;
}


inline XSubStr& XSubStr::operator = (const char* ys)
{
  assign(0, ys);
  return *this;
}

inline XSubStr& XSubStr::operator = (char ch)
{
  assign(0, &ch, 1);
  return *this;
}

inline XSubStr& XSubStr::operator = (const XStr& y)
{
  assign(y.rep, y.cChars(), y.length());
  return *this;
}

inline XSubStr& XSubStr::operator = (const XSubStr& y)
{
  assign(y.S.rep, y.cChars(), y.length());
  return *this;
}

// Zillions of cats...

inline void cat(const XStr& x, const XStr& y, XStr& r)
{
  r.rep = Scat(r.rep, x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const XStr& x, const XSubStr& y, XStr& r)
{
  r.rep = Scat(r.rep, x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const XStr& x, const char* y, XStr& r)
{
  r.rep = Scat(r.rep, x.cChars(), x.length(), y, -1);
}

inline void cat(const XStr& x, char y, XStr& r)
{
  r.rep = Scat(r.rep, x.cChars(), x.length(), &y, 1);
}

inline void cat(const XSubStr& x, const XStr& y, XStr& r)
{
  r.rep = Scat(r.rep, x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const XSubStr& x, const XSubStr& y, XStr& r)
{
  r.rep = Scat(r.rep, x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const XSubStr& x, const char* y, XStr& r)
{
  r.rep = Scat(r.rep, x.cChars(), x.length(), y, -1);
}

inline void cat(const XSubStr& x, char y, XStr& r)
{
  r.rep = Scat(r.rep, x.cChars(), x.length(), &y, 1);
}

inline void cat(const char* x, const XStr& y, XStr& r)
{
  r.rep = Scat(r.rep, x, -1, y.cChars(), y.length());
}

inline void cat(const char* x, const XSubStr& y, XStr& r)
{
  r.rep = Scat(r.rep, x, -1, y.cChars(), y.length());
}

inline void cat(const char* x, const char* y, XStr& r)
{
  r.rep = Scat(r.rep, x, -1, y, -1);
}

inline void cat(const char* x, char y, XStr& r)
{
  r.rep = Scat(r.rep, x, -1, &y, 1);
}

inline void cat(const XStr& a, const XStr& x, const XStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const XStr& a, const XStr& x, const XSubStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const XStr& a, const XStr& x, const char* y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x.cChars(), x.length(), y, -1);
}

inline void cat(const XStr& a, const XStr& x, char y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x.cChars(), x.length(), &y, 1);
}

inline void cat(const XStr& a, const XSubStr& x, const XStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const XStr& a, const XSubStr& x, const XSubStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const XStr& a, const XSubStr& x, const char* y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x.cChars(), x.length(), y, -1);
}

inline void cat(const XStr& a, const XSubStr& x, char y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x.cChars(), x.length(), &y, 1);
}

inline void cat(const XStr& a, const char* x, const XStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x, -1, y.cChars(), y.length());
}

inline void cat(const XStr& a, const char* x, const XSubStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x, -1, y.cChars(), y.length());
}

inline void cat(const XStr& a, const char* x, const char* y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x, -1, y, -1);
}

inline void cat(const XStr& a, const char* x, char y, XStr& r)
{
  r.rep = Scat(r.rep, a.cChars(), a.length(), x, -1, &y, 1);
}


inline void cat(const char* a, const XStr& x, const XStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const char* a, const XStr& x, const XSubStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const char* a, const XStr& x, const char* y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x.cChars(), x.length(), y, -1);
}

inline void cat(const char* a, const XStr& x, char y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x.cChars(), x.length(), &y, 1);
}

inline void cat(const char* a, const XSubStr& x, const XStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const char* a, const XSubStr& x, const XSubStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x.cChars(), x.length(), y.cChars(), y.length());
}

inline void cat(const char* a, const XSubStr& x, const char* y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x.cChars(), x.length(), y, -1);
}

inline void cat(const char* a, const XSubStr& x, char y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x.cChars(), x.length(), &y, 1);
}

inline void cat(const char* a, const char* x, const XStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, y.cChars(), y.length());
}

inline void cat(const char* a, const char* x, const XSubStr& y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, y.cChars(), y.length());
}

inline void cat(const char* a, const char* x, const char* y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, y, -1);
}

inline void cat(const char* a, const char* x, char y, XStr& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, &y, 1);
}


// operator versions

inline XStr& XStr::operator +=(const XStr& y)
{
  cat(*this, y, *this);
  return *this;
}

inline XStr& XStr::operator +=(const XSubStr& y)
{
  cat(*this, y, *this);
  return *this;
}

inline XStr& XStr::operator += (const char* y)
{
  cat(*this, y, *this);
  return *this;
}

inline XStr& XStr:: operator +=(char y)
{
  cat(*this, y, *this);
  return *this;
}

// constructive concatenation

#if defined(__GNUG__) && !defined(_G_NO_NRV)

inline XStr operator + (const XStr& x, const XStr& y) return r;
{
  cat(x, y, r);
}

inline XStr operator + (const XStr& x, const XSubStr& y) return r;
{
  cat(x, y, r);
}

inline XStr operator + (const XStr& x, const char* y) return r;
{
  cat(x, y, r);
}

inline XStr operator + (const XStr& x, char y) return r;
{
  cat(x, y, r);
}

inline XStr operator + (const XSubStr& x, const XStr& y) return r;
{
  cat(x, y, r);
}

inline XStr operator + (const XSubStr& x, const XSubStr& y) return r;
{
  cat(x, y, r);
}

inline XStr operator + (const XSubStr& x, const char* y) return r;
{
  cat(x, y, r);
}

inline XStr operator + (const XSubStr& x, char y) return r;
{
  cat(x, y, r);
}

inline XStr operator + (const char* x, const XStr& y) return r;
{
  cat(x, y, r);
}

inline XStr operator + (const char* x, const XSubStr& y) return r;
{
  cat(x, y, r);
}

inline XStr reverse(const XStr& x) return r;
{
  r.rep = Sreverse(x.rep, r.rep);
}

inline XStr upcase(const XStr& x) return r;
{
  r.rep = Supcase(x.rep, r.rep);
}

inline XStr downcase(const XStr& x) return r;
{
  r.rep = Sdowncase(x.rep, r.rep);
}

inline XStr capitalize(const XStr& x) return r;
{
  r.rep = Scapitalize(x.rep, r.rep);
}

#else /* NO_NRV */

inline XStr operator + (const XStr& x, const XStr& y)
{
  XStr r;  cat(x, y, r);  return r;
}

inline XStr operator + (const XStr& x, const XSubStr& y) 
{
  XStr r; cat(x, y, r); return r;
}

inline XStr operator + (const XStr& x, const char* y) 
{
  XStr r; cat(x, y, r); return r;
}

inline XStr operator + (const XStr& x, char y) 
{
  XStr r; cat(x, y, r); return r;
}

inline XStr operator + (const XSubStr& x, const XStr& y) 
{
  XStr r; cat(x, y, r); return r;
}

inline XStr operator + (const XSubStr& x, const XSubStr& y) 
{
  XStr r; cat(x, y, r); return r;
}

inline XStr operator + (const XSubStr& x, const char* y) 
{
  XStr r; cat(x, y, r); return r;
}

inline XStr operator + (const XSubStr& x, char y) 
{
  XStr r; cat(x, y, r); return r;
}

inline XStr operator + (const char* x, const XStr& y) 
{
  XStr r; cat(x, y, r); return r;
}

inline XStr operator + (const char* x, const XSubStr& y) 
{
  XStr r; cat(x, y, r); return r;
}

inline XStr reverse(const XStr& x) 
{
  XStr r; r.rep = Sreverse(x.rep, r.rep); return r;
}

inline XStr upcase(const XStr& x) 
{
  XStr r; r.rep = Supcase(x.rep, r.rep); return r;
}

inline XStr downcase(const XStr& x) 
{
  XStr r; r.rep = Sdowncase(x.rep, r.rep); return r;
}

inline XStr capitalize(const XStr& x) 
{
  XStr r; r.rep = Scapitalize(x.rep, r.rep); return r;
}

#endif

// prepend

inline void XStr::prepend(const XStr& y)
{
  rep = Sprepend(rep, y.cChars(), y.length());
}

inline void XStr::prepend(const char* y)
{
  rep = Sprepend(rep, y, -1); 
}

inline void XStr::prepend(char y)
{
  rep = Sprepend(rep, &y, 1); 
}

inline void XStr::prepend(const XSubStr& y)
{
  rep = Sprepend(rep, y.cChars(), y.length());
}

// misc transformations


inline void XStr::reverse()
{
  rep = Sreverse(rep, rep);
}


inline void XStr::upcase()
{
  rep = Supcase(rep, rep);
}


inline void XStr::downcase()
{
  rep = Sdowncase(rep, rep);
}


inline void XStr::capitalize()
{
  rep = Scapitalize(rep, rep);
}

// element extraction

inline char&  XStr::operator [] (int i) 
{ 
  if (i >= length()) error("invalid index");
  return rep->s[i];
}

inline const char&  XStr::operator [] (int i) const
{ 
  if (i >= length()) error("invalid index");
  return rep->s[i];
}

inline char  XStr::elem (int i) const
{ 
  if (i >= length()) error("invalid index");
  return rep->s[i];
}

inline char  XStr::firstchar() const
{ 
  return elem(0);
}

inline char  XStr::lastchar() const
{ 
  return elem(length() - 1);
}

// searching

inline int XStr::index(char c, int startpos) const
{
  return search(startpos, length(), c);
}

inline int XStr::index(const char* t, int startpos) const
{   
  return search(startpos, length(), t);
}

inline int XStr::index(const XStr& y, int startpos) const
{   
  return search(startpos, length(), y.cChars(), y.length());
}

inline int XStr::index(const XSubStr& y, int startpos) const
{   
  return search(startpos, length(), y.cChars(), y.length());
}

inline int XStr::index(const Regex& r, int startpos) const
{
  int unused;  return r.search(cChars(), length(), unused, startpos);
}

inline int XStr::contains(char c) const
{
  return search(0, length(), c) >= 0;
}

inline int XStr::contains(const char* t) const
{   
  return search(0, length(), t) >= 0;
}

inline int XStr::contains(const XStr& y) const
{   
  return search(0, length(), y.cChars(), y.length()) >= 0;
}

inline int XStr::contains(const XSubStr& y) const
{   
  return search(0, length(), y.cChars(), y.length()) >= 0;
}

inline int XStr::contains(char c, int p) const
{
  return match(p, length(), 0, &c, 1) >= 0;
}

inline int XStr::contains(const char* t, int p) const
{
  return match(p, length(), 0, t) >= 0;
}

inline int XStr::contains(const XStr& y, int p) const
{
  return match(p, length(), 0, y.cChars(), y.length()) >= 0;
}

inline int XStr::contains(const XSubStr& y, int p) const
{
  return match(p, length(), 0, y.cChars(), y.length()) >= 0;
}

inline int XStr::contains(const Regex& r) const
{
  int unused;  return r.search(cChars(), length(), unused, 0) >= 0;
}

inline int XStr::contains(const Regex& r, int p) const
{
  return r.match(cChars(), length(), p) >= 0;
}


inline int XStr::matches(const XSubStr& y, int p) const
{
  return match(p, length(), 1, y.cChars(), y.length()) >= 0;
}

inline int XStr::matches(const XStr& y, int p) const
{
  return match(p, length(), 1, y.cChars(), y.length()) >= 0;
}

inline int XStr::matches(const char* t, int p) const
{
  return match(p, length(), 1, t) >= 0;
}

inline int XStr::matches(char c, int p) const
{
  return match(p, length(), 1, &c, 1) >= 0;
}

inline int XStr::matches(const Regex& r, int p) const
{
  int l = (p < 0)? -p : length() - p;
  return r.match(cChars(), length(), p) == l;
}


inline int XSubStr::contains(const char* t) const
{   
  return S.search(pos, pos+len, t) >= 0;
}

inline int XSubStr::contains(const XStr& y) const
{   
  return S.search(pos, pos+len, y.cChars(), y.length()) >= 0;
}

inline int XSubStr::contains(const XSubStr&  y) const
{   
  return S.search(pos, pos+len, y.cChars(), y.length()) >= 0;
}

inline int XSubStr::contains(char c) const
{
  return S.search(pos, pos+len, c) >= 0;
}

inline int XSubStr::contains(const Regex& r) const
{
  int unused;  return r.search(cChars(), len, unused, 0) >= 0;
}

inline int XSubStr::matches(const Regex& r) const
{
  return r.match(cChars(), len, 0) == len;
}


inline int XStr::gsub(const XStr& pat, const XStr& r)
{
  return _gsub(pat.cChars(), pat.length(), r.cChars(), r.length());
}

inline int XStr::gsub(const XSubStr&  pat, const XStr& r)
{
  return _gsub(pat.cChars(), pat.length(), r.cChars(), r.length());
}

inline int XStr::gsub(const Regex& pat, const XStr& r)
{
  return _gsub(pat, r.cChars(), r.length());
}

inline int XStr::gsub(const char* pat, const XStr& r)
{
  return _gsub(pat, -1, r.cChars(), r.length());
}

inline int XStr::gsub(const char* pat, const char* r)
{
  return _gsub(pat, -1, r, -1);
}



inline  ostream& operator<<(ostream& s, const XStr& x)
{
   s << x.cChars(); return s;
}

// a zillion comparison operators

inline int operator==(const XStr& x, const XStr& y) 
{
  return compare(x, y) == 0; 
}

inline int operator!=(const XStr& x, const XStr& y)
{
  return compare(x, y) != 0; 
}

inline int operator>(const XStr& x, const XStr& y)
{
  return compare(x, y) > 0; 
}

inline int operator>=(const XStr& x, const XStr& y)
{
  return compare(x, y) >= 0; 
}

inline int operator<(const XStr& x, const XStr& y)
{
  return compare(x, y) < 0; 
}

inline int operator<=(const XStr& x, const XStr& y)
{
  return compare(x, y) <= 0; 
}

inline int operator==(const XStr& x, const XSubStr&  y) 
{
  return compare(x, y) == 0; 
}

inline int operator!=(const XStr& x, const XSubStr&  y)
{
  return compare(x, y) != 0; 
}

inline int operator>(const XStr& x, const XSubStr&  y)      
{
  return compare(x, y) > 0; 
}

inline int operator>=(const XStr& x, const XSubStr&  y)
{
  return compare(x, y) >= 0; 
}

inline int operator<(const XStr& x, const XSubStr&  y) 
{
  return compare(x, y) < 0; 
}

inline int operator<=(const XStr& x, const XSubStr&  y)
{
  return compare(x, y) <= 0; 
}

inline int operator==(const XStr& x, const char* t) 
{
  return compare(x, t) == 0; 
}

inline int operator!=(const XStr& x, const char* t) 
{
  return compare(x, t) != 0; 
}

inline int operator>(const XStr& x, const char* t)  
{
  return compare(x, t) > 0; 
}

inline int operator>=(const XStr& x, const char* t) 
{
  return compare(x, t) >= 0; 
}

inline int operator<(const XStr& x, const char* t)  
{
  return compare(x, t) < 0; 
}

inline int operator<=(const XStr& x, const char* t) 
{
  return compare(x, t) <= 0; 
}

inline int operator==(const XSubStr& x, const XStr& y) 
{
  return compare(y, x) == 0; 
}

inline int operator!=(const XSubStr& x, const XStr& y)
{
  return compare(y, x) != 0;
}

inline int operator>(const XSubStr& x, const XStr& y)      
{
  return compare(y, x) < 0;
}

inline int operator>=(const XSubStr& x, const XStr& y)     
{
  return compare(y, x) <= 0;
}

inline int operator<(const XSubStr& x, const XStr& y)      
{
  return compare(y, x) > 0;
}

inline int operator<=(const XSubStr& x, const XStr& y)     
{
  return compare(y, x) >= 0;
}

inline int operator==(const XSubStr& x, const XSubStr&  y) 
{
  return compare(x, y) == 0; 
}

inline int operator!=(const XSubStr& x, const XSubStr&  y)
{
  return compare(x, y) != 0;
}

inline int operator>(const XSubStr& x, const XSubStr&  y)      
{
  return compare(x, y) > 0;
}

inline int operator>=(const XSubStr& x, const XSubStr&  y)
{
  return compare(x, y) >= 0;
}

inline int operator<(const XSubStr& x, const XSubStr&  y) 
{
  return compare(x, y) < 0;
}

inline int operator<=(const XSubStr& x, const XSubStr&  y)
{
  return compare(x, y) <= 0;
}

inline int operator==(const XSubStr& x, const char* t) 
{
  return compare(x, t) == 0; 
}

inline int operator!=(const XSubStr& x, const char* t) 
{
  return compare(x, t) != 0;
}

inline int operator>(const XSubStr& x, const char* t)  
{
  return compare(x, t) > 0; 
}

inline int operator>=(const XSubStr& x, const char* t) 
{
  return compare(x, t) >= 0; 
}

inline int operator<(const XSubStr& x, const char* t)  
{
  return compare(x, t) < 0; 
}

inline int operator<=(const XSubStr& x, const char* t) 
{
  return compare(x, t) <= 0; 
}


// a helper needed by at, before, etc.

inline XSubStr XStr::_substr(int first, int l)
{
  if (first < 0 || (first + l) > length() )
    return XSubStr(_nilString, 0, 0) ;
  else 
    return XSubStr(*this, first, l);
}

#endif



